/*
    Trabalho Prtico I - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Implementacao do filtro dos k vizinhos mais proximos

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0

typedef struct T_ValorDiferenca
{
    int valor;
    int diferenca;
} ValorDiferenca;


void kVizinhos(int *imagem, int *imagemSaida, int nl, int nc);
int maiorValor(ValorDiferenca *a, ValorDiferenca *b);

static int TAMJANELA;
static int K;

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn;
    int *imagemEntrada;
    int *imagemSaida;

    printf("\nFiltro dos K-VIZINHOS by <Anderson Rocha, Leyza Dorini>\n");
    if (argc == 5)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);

        info_imagem (argv[1], nl, nc, mn);
        aloca_memo (&imagemSaida, nl, nc);

        TAMJANELA = atoi(argv[3]);
        K = atoi(argv[4]);

        if(TAMJANELA < nl && TAMJANELA < nc)
        {
            if(K < (TAMJANELA * TAMJANELA))
            {
                kVizinhos(imagemEntrada, imagemSaida, nl, nc);

                if(imagemEntrada != null)
                {
                    grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
                    printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
                }
            }
            else
            {
                tratarErro(1);
            }
        }
        else
        {
            tratarErro(2);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(4);
    }

    printf("\n");
    return(0);
}

/*
    imagem = imagem de entrada
    imagemSaida = imagem esperada como resultado
    nl = numero de linhas
    nc = numero de colunas
*/
void kVizinhos(int *imagem, int *imagemSaida, int nl, int nc)
{
    // coluna, linha, nro colunas da janela, nro linhas
    int c, l, i, j, k;
    // Janela que ir percorrer a imagem selecionada
    int janela[TAMJANELA][TAMJANELA];
    // Numero de linhas que podem ser percorridas pela janela
    int x;
    // Numero de colunas que podem ser percorridas pela janela
    int nj;
    // Referencia ao elemento central de uma janela
    int elementoCentral = 0;
    // Auxiliar para se achar a posicao do elemento central de uma janela em relao
    //  sua posio absoluta na imagem
    int contador = 0;
    // Reflete qual o elemento da janela passada deve ser alterado com o
    // valor de k passado
    int posicao = (TAMJANELA * TAMJANELA)/ 2 + 1;
    // Soma e media dos k pixels selecinados como vizinhos
    int soma, media;
    // Guarda em cada tupla <v,d> o valor do pixel e a sua diferenca em relacao ao
    // elemento central da janela em questao
    ValorDiferenca vd[TAMJANELA * TAMJANELA];

    nj = 0;
    x = 0;

    // Achando todas as janelas TxT (3x3, 5x5, 7x7...) possveis dentro
    // da imagem
    for(x = 0; x < nl - TAMJANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAMJANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAMJANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAMJANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }

                k = (k + nc);
                l++;
            }

            // Acumular em VD os valores da janela[i][j] e as diferencas em relaao ao elemento central
            for(i = 0; i < TAMJANELA; i++)
            {
                for(j = 0; j < TAMJANELA; j++)
                {
                    vd[i * TAMJANELA + j].diferenca = abs(janela[TAMJANELA/2][TAMJANELA/2] - janela[i][j]);
                    vd[i * TAMJANELA + j].valor = janela[i][j];
                }
            }

            // Inserindo INFINITO no elemento do MEIO para que ele no entre na comparao. Explicao: dado X, o elemento
            // mais proximo de X  ele mesmo. Logo X no pode entrar na amostragem.
            vd[(TAMJANELA * TAMJANELA)/2].valor = INFINITO;
            vd[(TAMJANELA * TAMJANELA)/2].diferenca = INFINITO;

            // Ordenando a sequencia de <v,d>'s de modo que seja possvel escolher quem so os mais prximos do elemento
            // central que agora est como INFINITO para no ser selecionado na amostragem.
            qsort(vd, (TAMJANELA * TAMJANELA), sizeof(ValorDiferenca), (int (*) (const void *, const void *)) maiorValor);

            // A soma dos kVizinhos a serem selecionados deve comear com zero
            soma = 0;

            // Calculando a soma dos kVizinhos selecionados
            for(i = 0; i < K; i++)
                soma += vd[i].valor;

            media = soma / K;
            // Atualizando o elemento na imagem de resultado
            *(imagemSaida + elementoCentral) = media;
        }
    }
}

// Calcula quem  o maior elemento de uma tupla T1 = <v1, d1> e T2 = <v2, d2>. A ordem est relacionada ao menor "d"
int maiorValor(ValorDiferenca *a, ValorDiferenca *b)
{
    if(a->diferenca > b->diferenca)
        return (1);
    else if(a->diferenca == b->diferenca)
        return (0);
    return (-1);
}
