/*
    Trabalho Prtico I - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Implementacao do filtro de NAGAO

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0
// Tamanho da janela
#define TAMJANELA 5

typedef struct T_MediaVariancia
{
    int media;
    int variancia;
} MediaVariancia;

void nagao(int *imagem, int *imagemSaida, int nl, int nc);
MediaVariancia taxaDispersao(int janela[][TAMJANELA], int li, int ci, int lf, int cf, int x1, int y1, int x2, int y2, int chave);

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn;
    int *imagemEntrada;
    int *imagemSaida;

    printf("\nFiltro de NAGAO by <Anderson Rocha, Leyza Dorini>\n");

    if (argc == 3)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);

        info_imagem (argv[1], nl, nc, mn);


        aloca_memo (&imagemSaida, nl, nc);

        nagao(imagemEntrada, imagemSaida, nl, nc);

        if(imagemEntrada != null)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }
        else
        {
            tratarErro(3);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(2);
    }

    printf("\n");
    return(0);
}

/*
    imagem = imagem armazenada
    nl = numero de linhas
    nc = numero de colunas
    tmx = tamanho da janela em x
    tmy = tamanho da janela em y
*/
void nagao(int *imagem, int *imagemSaida, int nl, int nc)
{
    // coluna,  linha, numero de janelas, nro colunas da janela, nro linhas
    int c, l, nj, i, j, k;
    int janela[TAMJANELA][TAMJANELA];
    int x;
    int menorDispersao;
    int elementoCentral = 0;
    MediaVariancia dispersoes[9];
    int mediaMenorDispersao = 0;
    int contador = 0;
    // Usado para receber a mdia calculada pela funo de clculo de disperses
    int *media = null;

    nj = 0;
    x = 0;

    for(x = 0; x < nl - TAMJANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAMJANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAMJANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAMJANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == 13)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            // Chamada para o clculo das dispersoes das regioes proposta por NAGAO
            // 4 Hexagonos
            dispersoes[0] = taxaDispersao(janela, 0, 0, 2, 2, 2, 0, 0, 2, 1);
            dispersoes[1] = taxaDispersao(janela, 0, 2, 2, 4, 0, 2, 2, 4, 1);
            dispersoes[2] = taxaDispersao(janela, 2, 0, 4, 2, 2, 0, 4, 2, 1);
            dispersoes[3] = taxaDispersao(janela, 2, 2, 4, 4, 4, 2, 2, 4, 1);
            // 4 Pentagonos
            dispersoes[4] = taxaDispersao(janela, 0, 1, 2, 3, 2, 1, 2, 3, 1);
            dispersoes[5] = taxaDispersao(janela, 1, 0, 3, 2, 1, 2, 3, 2, 1);
            dispersoes[6] = taxaDispersao(janela, 2, 1, 4, 3, 2, 1, 2, 3, 1);
            dispersoes[7] = taxaDispersao(janela, 1, 2, 3, 4, 1, 2, 3, 2, 1);
            // 1 Retangulo
            dispersoes[8] = taxaDispersao(janela, 1, 1, 3, 3, 0, 0, 0, 0, 0);

            menorDispersao = dispersoes[0].variancia;
            mediaMenorDispersao = dispersoes[0].media;

            for(i = 1; i < 9; i++)
            {
                if(dispersoes[i].variancia < menorDispersao)
                {
                    menorDispersao = dispersoes[i].variancia;
                    mediaMenorDispersao = dispersoes[i].media;
                }
            }

            *(imagemSaida + elementoCentral) = mediaMenorDispersao;
        }
    }
}

// Este indice de disperso e calculada pela variancia. Poderia ser desvio-padro,
// amplitude, amplitude inter-quartil, entre outros.
// (x1, y1) e (x2, y2) so os pontos a serem excluidos numa janela 3x3 de modo a formarem pentagonos ou hexagonos
// Caso tipo seja ZERO a figura a ser tratada  um retangulo. Desta forma, no se tem (x1, y1) e (x2, y2)
// como pontos a serem excludos
MediaVariancia taxaDispersao(int janela[][TAMJANELA], int li, int ci, int lf, int cf, int x1, int y1, int x2, int y2, int chave)
{
    int i, j, soma, nroElementos, media, variancia;
    // indice dos excluidos
    int iex;
    MediaVariancia mv;

    variancia = 0;
    soma = 0;
    media = 0;

    // Calcula a soma
    for(i = li; i <= lf; i++)
    {
        for(j = ci; j <= cf; j++)
        {
            soma = soma + janela[i][j];
        }
    }

    if(chave)
    {
        // Qualquer pentagano ou hexagono tem 7 elementos
        // 6 de borda e 1 interno
        nroElementos = 7;
        soma = (soma - janela[x1][y1] - janela[x2][y2]);
    }
    else
    {
        // Um retangulo tem 9 elementos
        nroElementos = 9;
    }

    media = (soma / nroElementos);

    for(i = li; i <= lf; i++)
    {
        for(j = ci; j <= cf; j++)
        {
            variancia += pow((janela[i][j] - media), 2);
        }
    }

    if(chave)
    {
        variancia -= pow((janela[x1][y1] - media), 2);
        variancia -= pow((janela[x2][y2] - media), 2);
    }

    variancia = variancia / (nroElementos);

    mv.media = media;
    mv.variancia = variancia;

    return mv;
}
