/*
    Trabalho Prtico I - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Implementacao do filtro da Ordem K

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0

void ordemK(int *imagem, int *imagemSaida, int nl, int nc);
int maiorValor(int *a, int *b);

static int TAMJANELA;
static int K;

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn;
    int *imagemEntrada;
    int *imagemSaida;

    printf("\nFiltro da ORDEM-K by <Anderson Rocha, Leyza Dorini>\n");
    if (argc == 5)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);

        info_imagem (argv[1], nl, nc, mn);
        aloca_memo (&imagemSaida, nl, nc);

        TAMJANELA = atoi(argv[3]);
        K = atoi(argv[4]);

        if(TAMJANELA < nl && TAMJANELA < nc)
        {
            if(K < (TAMJANELA * TAMJANELA))
            {
                ordemK(imagemEntrada, imagemSaida, nl, nc);

                if(imagemEntrada != null)
                {
                    grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
                    printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
                }
            }
            else
            {
                tratarErro(1);
            }
        }
        else
        {
            tratarErro(2);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(3);;
    }

    printf("\n");
    return(0);
}

/*
    imagem = imagem armazenada
    nl = numero de linhas
    nc = numero de colunas
    tmx = tamanho da janela em x
    tmy = tamanho da janela em y
*/
void ordemK(int *imagem, int *imagemSaida, int nl, int nc)
{
    // coluna,  linha, numero de janelas, nro colunas da janela, nro linhas
    int c, l, nj, i, j, k;
    int janela[TAMJANELA][TAMJANELA];
    int x;
    int menorDispersao;
    int elementoCentral = 0;
    int contador = 0;
    // Reflete qual o elemento da janela passada deve ser alterado com o
    // valor de k passado
    int posicao = (TAMJANELA * TAMJANELA)/ 2 + 1;

    nj = 0;
    x = 0;

    for(x = 0; x < nl - TAMJANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAMJANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAMJANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAMJANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            // Ordenando a janela construda
            qsort(janela, (TAMJANELA * TAMJANELA), sizeof(int), (int (*) (const void *, const void *)) maiorValor);
              *(imagemSaida + elementoCentral) = janela[K / TAMJANELA][K % TAMJANELA];
        }
    }
}

int maiorValor(int *a, int *b)
{
    if(*a > *b)
        return (1);
    else if(*a == *b)
        return (0);
    return (-1);
}
