/*
    Trabalho Prtico I - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Implementacao dos operadores de Prewitt para deteco de contorno

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0

void prewitt(int *imagem, int *imagemSaida, int nl, int nc);

static int TAMJANELA = 3;

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn;
    int *imagemEntrada;
    int *imagemSaida;

    printf("\nOperadores de Prewitt, by <Anderson Rocha, Leyza Dorini>\n");
    if (argc == 3)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);

        info_imagem (argv[1], nl, nc, mn);
        aloca_memo (&imagemSaida, nl, nc);

        if(TAMJANELA < nl && TAMJANELA < nc)
        {
    		prewitt(imagemEntrada, imagemSaida, nl, nc);

	    	if(imagemEntrada != null)
		    {
			    grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
			    printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
		    }
        }
        else
        {
            tratarErro(2);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(6);;
    }

    printf("\n");
    return(0);
}

/*
    imagem = imagem armazenada
    nl = numero de linhas
    nc = numero de colunas
    tmx = tamanho da janela em x
    tmy = tamanho da janela em y
*/
void prewitt(int *imagem, int *imagemSaida, int nl, int nc)
{
    // coluna,  linha, numero de janelas, nro colunas da janela, nro linhas
    int c, l, nj, i, j, k;
    int janela[TAMJANELA][TAMJANELA];
    int x;
    int valOpX, valOpY;
    int elementoCentral = 0;
    int contador = 0;
    int aux;
    // Reflete qual o elemento da janela passada deve ser alterado com o
    // valor de k passado
    int posicao = (TAMJANELA * TAMJANELA)/ 2 + 1;
    // Operador para a vertical
    int operadorX[3][3] = {-1, -1, -1, 0, 0, 0, 1, 1, 1};
    // Operador para a horizontal
    int operadorY[3][3] = {-1, 0, 1, -1, 0, 1, -1, 0, 1};

    printf("Criou operadores\n");

    nj = 0;
    x = 0;
    aux=0;

    for(x = 0; x < nl - TAMJANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAMJANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAMJANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAMJANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            // Aplica os operadores separadamente
            valOpX = aplicarOperador(janela, operadorX);
            valOpY = aplicarOperador(janela, operadorY);

            // Faz um ou lgico de modo que se possa detectar tanto os contornos
            // em X quanto em Y. A operao absoluta  necessria porque algumas vezes
            // o operador de Prewitt gera resultados negativos que devem ser desconsiderados
            // dado que tudo que  mapeado de -128 a 127 pode ser repassado para 0 a 255            
            aux = abs(valOpX + valOpY);
            //agora  realizada uma comparao da sada da soma das mscaras com um limiar,
            //com o intuito de criar uma mscara binarizada, que possua valor 255 nas regies
            //de alta freqncia (bordas). Esta mscara ser utilizada para clculo do RMS High.
            //Para calcular o RMS Low, utiliza-se a operao de negativo nesta imagem.
            //O limiar foi definido de forma emprica.
            //Se no fosse feita esta limiarizao, a imagem de sada receberia:
            //*(imagemSaida + elementoCentral) = abs(valOpX + valOpY);
            if(aux>65)
            {    
                *(imagemSaida + elementoCentral) = 255;            
            }
            else
            {
                *(imagemSaida + elementoCentral) = 0;
            }                    
        }
    }
}

// Aplica efetivamente um operador no eixo do X ou Y dependendo do que vier como
// parmetro
int aplicarOperador(int janela[][TAMJANELA], int operador[][TAMJANELA])
{
    int i, j, soma, aux;

    soma = 0;

    for(i = 0; i < TAMJANELA; i++)
    {
        for(j = 0; j < TAMJANELA; j++)
        {            
           soma += janela[i][j] * operador[i][j];
            //printf("A soma eh: %d\n", soma);
        }
    }

    //printf("A soma eh: %d\n", soma);
    return soma;    
}
